/*******************************************************************************
 * Copyright 2010 Seapine Software, Inc. All rights reserved.
 * 
 * IMPORTANT:  This Software is supplied to you by Seapine Software, Inc. (Seapine)
 * in consideration of your agreement to the following terms, and your use, installation,
 * modification or redistribution of this Software constitutes acceptance of these terms.
 * If you do not agree with these terms, please do not use, install, modify or redistribute
 * this Software.
 * 
 * In consideration of your agreement to abide by the following terms, and subject to these
 * terms, Seapine grants you a personal, non-exclusive license, under Seapine copyrights in
 * this original Seapine software (the Software), to use, reproduce, modify and redistribute
 * the Software, with or without modifications, in source and/or binary forms.
 * Neither the name, trademarks, service marks or logos of Seapine, Inc. may be used to endorse
 * or promote products derived from the Software without specific prior written permission from
 * Seapine.  Except as expressly stated in this notice, no other rights or licenses, express or
 * implied, are granted by Seapine herein, including but not limited to any patent rights that
 * may be infringed by your derivative works or by other works in which the Software may be
 * incorporated.
 * 
 * The Software is provided by Seapine on an "AS IS" basis.  SEAPINE MAKES NO WARRANTIES,
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF NON-INFRINGEMENT,
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, REGARDING THE SOFTWARE OR ITS USE AND
 * OPERATION ALONE OR IN COMBINATION WITH YOUR PRODUCTS. 
 * 
 * IN NO EVENT SHALL SEAPINE BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) ARISING IN ANY WAY OUT OF THE USE,
 * REPRODUCTION, MODIFICATION AND/OR DISTRIBUTION OF THE SOFTWARE, HOWEVER CAUSED AND WHETHER
 * UNDER THEORY OF CONTRACT, TORT (INCLUDING NEGLIGENCE), STRICT LIABILITY OR OTHERWISE, EVEN
 * IF SEAPINE HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************/

#include <stdlib.h>
#include <string>
#include <time.h>
#include <math.h>
#include <fstream>
#include <iostream>

#include "SSCMIssuePlugin.h"
#include "XmlRpc.h"
using namespace XmlRpc;
using namespace std;

// global variables
std::string currentUser;
std::string currentToken;
std::string cookieFilePath;
std::string configFilePath;

// required metadata 
extern "C" {
   IPExport char SSCMIP_PluginProvider[]  = "Bugzilla";
   IPExport char SSCMIP_PluginUUID[]      = "517f1c80-85f9-11df-a4ee-0800200c9a66";
   IPExport char SSCMIP_PluginAuthors[]   = "Seapine Software";
   IPExport char SSCMIP_PluginNotes[]     = "This plugin allows Surround SCM to process attachments for Bugzilla.";
   IPExport char SSCMIP_PluginVersion[]   = "1.0.0";
   IPExport int SSCMIP_APIVersion         = _SSCMIP_APIVersion_;
}

// filenames for cookies and connections file
#ifdef WIN32
const char* myCookieFile = "\\cookies.txt";
const char* myConfigFile = "\\BugzillaConnections.txt";
#else 
const char* myCookieFile = "/cookies.txt";
const char* myConfigFile = "/BugzillaConnections.txt";
#endif

///// Additional Function Declarations
char* CopyAndAllocStr(const char *pStr);
char* CopyAndAllocFormatStr(const char *pStr, int num);
char* CopyAndAllocFormatStr(const char *pStr,  const char* num);
char* ValToStr(XmlRpcValue value);
char* TimeToStr(XmlRpcValue value);
SSCMIPConnectionDetails** AllocConnectionDetailsList(uint numOfItems);
SSCMIPCustomField** AllocCustomFieldList(uint numOfItems);
SSCMIPCustomFieldValue** AllocCustomFieldValueList(uint numOfItems);
SSCMIPIssue** AllocIssueList(uint numOfItems);
SSCMIPFailedIssueAttachment** AllocFailedIssueAttachment(uint numOfItems);

SSCMIPResult bzLogin(XmlRpcClient c, 
                     const SSCMIPConnectionContext* info, 
                     char** errorMsg);
SSCMIPResult handleErrors(SSCMIPResult status, 
                          const char* operation, 
                          XmlRpcValue result,  
                          char** errorString);
SSCMIPResult bzExecute(XmlRpcClient* c,
                       const SSCMIPConnectionContext* info, 
                       const char* function, 
                       XmlRpcValue parameters, 
                       XmlRpcValue* sucess);
BOOL expired (ifstream* cookie);
SSCMIPResult readBugzillaConnections(XmlRpcValue* connectionList, 
                                     char** errorString);
SSCMIPResult getConnectionInfo(XmlRpcValue* info, 
                               const char* connectionName, 
                               char** errorString);

BOOL rightUser(SSCMIPConnectionContext context);

//////////////////////////////////////////////
// Export Functions
//////////////////////////////////////////////

// Initialize the plugin, called when the plugin is loaded into memory
IPExport SSCMIPResult Initialize(const char *pDirectoryPath, 
                                 char **pErrorString)
{

   // stores the directory path as a global variable
   cookieFilePath = string(pDirectoryPath).append(myCookieFile);
   configFilePath = string(pDirectoryPath).append(myConfigFile);

   // sets the current user to an empty string
   currentUser.clear();

   // Return OK
   *pErrorString = NULL;
   return SSCMIP_OK;
}

// Uninitalize the plugin, called when the plugin is being removed from memory
IPExport void Uninitialize()
{
   // remove the cookies file
   remove(cookieFilePath.c_str());
}

// Provides a list of server connections
 IPExport SSCMIPResult GetConnectionList(const SSCMIPConnectionContext *context,
                                         SSCMIPConnectionDetails **pConnections[],
                                         uint *numOfConnections,
                                         char **pErrorString)

{
   *pErrorString = NULL;
   XmlRpcValue connections;
   // read the connections file
   SSCMIPResult result = readBugzillaConnections(&connections, pErrorString);
   if (!result)
   {
      // find the size and allocate the list
      uint connectionCount = connections.size();
      SSCMIPConnectionDetails** detailList = AllocConnectionDetailsList(connectionCount);
      // loop through and copy the names to the list 
      for (uint i=0; i<connectionCount; i++)
      {
         detailList[i]->pConnectionName = ValToStr(connections[i]["name"]);
         detailList[i]->bConnectionSupportsBulkFixEvents = true;
      }
      // assign the pointers
      *pConnections = detailList;
      *numOfConnections = connectionCount;
   }
   return result;
}

// Provides a list of filters, optional. 
// (*pFilters) should be NULL and (*numOfFilters) should be 0 if not implemented.
IPExport SSCMIPResult GetFilterList(const SSCMIPConnectionContext *context,
                                    char **pFilters[],
                                    uint *numOfFilters,
                                    char **pErrorString)
{
   // three default filters
   uint filterCount = 3;
   char** filters = new char*[filterCount];
   *numOfFilters = filterCount;
   
   filters[0] = CopyAndAllocStr("Assigned To Me");
   filters[1] = CopyAndAllocStr("Created By Me");
   filters[2] = CopyAndAllocStr("Not Filtered");

   *pFilters = filters;
   *pErrorString = NULL;
   return SSCMIP_OK;
}

// Provides a list of issues
// numericIssueIds should be true only if all id's are numeric only. This allows for numeric sorting on the client side.
 IPExport SSCMIPResult GetIssueList(const SSCMIPConnectionContext *context,
                                   const char *pFilter,
                                   SSCMIPIssue **issues[],
                                   uint *numOfIssues,
                                   BOOL *numericIssueIds,
                                   char **pErrorString)
{
   BOOL numeric = true;
   XmlRpcValue connection, params, getResult, comResult, getParams, comParams, ids;
   SSCMIPResult result = getConnectionInfo(&connection, context->pConnection, pErrorString);

   if (!result)
   {
      // setup XmlRpcClient according to selected connection
      XmlRpcClient Client (ValToStr(connection["server_url"]), connection["port"], ValToStr(connection["web_service_extension"]));
      Client.setCookieFilePath(cookieFilePath);
      currentToken.clear();

      // set parameters based on product and filters
      getParams["product"] = connection["product"];
      if ( strcmp(pFilter, "Assigned To Me" ) == 0 )
      {
         getParams["assigned_to"] = context->pUsername;
      }
      else if ( strcmp(pFilter, "Created By Me" ) == 0 )
      {
         getParams["reporter"] = context->pUsername;
      }
      params[0] = getParams;
      
      // search for bugs by criteria
      result = bzExecute(&Client, context, "Bug.search", params, &getResult);

      if (!result && !getResult.hasMember("faultCode"))
      {
         // allocate list of issues
         uint i, length = getResult["bugs"].size();

         // if no bugs pass the filter, don't continue, or Bug.comment will whine
         if (length > 0)
         {
            // make a list of bug ids, pass to comments function
            for (i=0; i<length; i++) 
            {
               ids[i]=getResult["bugs"][i]["id"];
            }
            comParams["ids"]= ids;
            params[0]= comParams;
            result = bzExecute(&Client, context, "Bug.comments", params, &comResult);

            if(!result && !comResult.hasMember("faultCode"))
            {
               SSCMIPIssue** myIssues = AllocIssueList(length);
               // loop through and assign values
               for (i=0; i<length; i++)
               {
                  myIssues[i]->pIssueId = ValToStr(getResult["bugs"][i]["id"]);
                  myIssues[i]->pSummary = ValToStr(getResult["bugs"][i]["summary"]);
                  myIssues[i]->pDescription = ValToStr(comResult["bugs"][(std::string)
                     getResult["bugs"][i]["id"]]["comments"][0]["text"]);
                  myIssues[i]->pStatus = ValToStr(getResult["bugs"][i]["status"]);
                  myIssues[i]->pType = ValToStr(getResult["bugs"][i]["severity"]);
                  myIssues[i]->bUserCanAttach = true;
                  myIssues[i]->bUserCanFix = true;
               }
               // set pointers
               *issues = myIssues;
               *numOfIssues = length;
               *numericIssueIds = numeric;
            }
            else 
            {
               result = handleErrors(result, "Get Comments", comResult, pErrorString);
            }
         }
      }
      else 
      {
         result = handleErrors(result, "Bug Search", getResult, pErrorString);
      }
   }
   if (result) 
   {
      result = SSCMIP_ERROR;
   }
   return result;
}

// Provides the Custom Fields for an issue
IPExport SSCMIPResult GetIssueDetails(const SSCMIPConnectionContext *context,
                                      const char *pIssueId,
                                      SSCMIPIssue **pIssue,
                                      SSCMIPCustomField **customFields[],
                                      uint *numOfCustomFields,
                                      char **pErrorString)
{
   XmlRpcValue connection, params, verResult, getResult, comResult, fieldResult, 
      bzParams, fieldNames, bugNames;

   SSCMIPResult result = getConnectionInfo(&connection, context->pConnection, pErrorString);

   if (!result)
   {
      // setup XmlRpcClient according to selected connection
      XmlRpcClient Client (ValToStr(connection["server_url"]), connection["port"], 
         ValToStr(connection["web_service_extension"]));
      Client.setCookieFilePath(cookieFilePath);
      // check bugzilla's version number, to make sure we are calling the right function
      result = bzExecute(&Client, context, "Bugzilla.version", params, &verResult);
   
      if (!result && !verResult.hasMember("faultCode"))
      {
         // pass issue id to bug.get
         bzParams["ids"][0] = pIssueId;
         bzParams["token"] = currentToken;
         params[0] = bzParams;
         
         if ((std::string)verResult["version"] >= "5.0")
            result = bzExecute(&Client, context, "Bug.get", params, &getResult);
         else
            result = bzExecute(&Client, context, "Bug.get_bugs", params, &getResult);

         if (!result && !getResult.hasMember("faultCode"))
         {
            uint i, length = getResult["bugs"].size();

            // if no bugs pass the filter, don't continue, or Bug.comment will whine
            XmlRpcValue bugCommentParams;
            if (length > 0)
            {               
               // make a list of bug ids, pass to comments function
               for (i = 0; i < length; i++)
               {
                  bugCommentParams["ids"][i] = getResult["bugs"][i]["id"];
               }
            }

            // get comments
            result = bzExecute(&Client, context, "Bug.comments", bugCommentParams, &comResult);

            if (!result && !comResult.hasMember("faultCode"))
            {
               // Issue Info
               SSCMIPIssue* myIssue = new SSCMIPIssue;
               myIssue->pIssueId = ValToStr(getResult["bugs"][0]["id"]);
               myIssue->pSummary = ValToStr(getResult["bugs"][0]["summary"]);
               myIssue->pDescription = 
                  ValToStr(comResult["bugs"][(std::string) myIssue->pIssueId]["comments"][0]["text"]);
               myIssue->pStatus = ValToStr(getResult["bugs"][0]["status"]);
               myIssue->pType = ValToStr(getResult["bugs"][0]["severity"]);
               myIssue->bUserCanAttach = true;
               myIssue->bUserCanFix = true;
               *pIssue = myIssue;
               
               // if the bug is open, don't expect a resolution
               uint i, k=0, cfCount;
               if (getResult["bugs"][0]["is_open"]) i=1;
               else  i=0; 
               cfCount = 9-i;

               SSCMIPCustomField** fieldList = AllocCustomFieldList(cfCount);

               bugNames[0] = "resolution"; 
               bugNames[1] = "product" ;  
               bugNames[2] = "component" ;
               bugNames[3] = "priority"; 
               bugNames[4] = "severity"; 
               bugNames[5] = "alias";
               bugNames[6] = "assigned_to";
               bugNames[7] = "last_change_time";
               bugNames[8] = "creation_time";

               fieldNames["names"][8] = "creation_ts";

               SSCMIPCustomFieldType fieldTypes[] = {eDropdown, eDropdown, eDropdown, eDropdown, 
                  eDropdown, eTextfield, eTextfield, eDateTime, eDateTime};

               // newer version, lets us call bug.fields
               if ( (std::string)verResult["version"] >= "3.6" )
               {
                  bugNames[4] = "bug_severity";

                  fieldNames["names"][0] = "resolution"; 
                  fieldNames["names"][1] = "product" ;  
                  fieldNames["names"][2] = "component" ;
                  fieldNames["names"][3] = "priority"; 
                  fieldNames["names"][4] = "bug_severity"; 
                  fieldNames["names"][5] = "alias";
                  fieldNames["names"][6] = "assigned_to";
                  fieldNames["names"][7] = "delta_ts";

                  params[0] = fieldNames;
                  result = bzExecute(&Client, context, "Bug.fields", fieldNames, &fieldResult);

                  if(!result && !fieldResult.hasMember("faultCode"))
                  {
                     // if bug is closed, set resolution
                     if (!getResult["bugs"][0]["is_open"])
                     {
                        fieldList[k]->pDisplayName = ValToStr(fieldResult["fields"][0]["display_name"]);
                        fieldList[k]->pFieldId = ValToStr(fieldNames["names"][0]);
                        fieldList[k]->eType = fieldTypes[0];
                        fieldList[k]->numOfFieldValues = 1;
                        fieldList[k]->pCustomFieldValues = AllocCustomFieldValueList(1); 
                        // if resolution is duplicate, add the dupe #
                        if( (std::string)getResult["bugs"][0]["resolution"] == "DUPLICATE")
                        {
                           fieldList[k]->pCustomFieldValues[0]->pDisplayValue = 
                              ValToStr( (std::string)getResult["bugs"][0]["resolution"] + 
                              " of bug # " + (std::string)getResult["bugs"][0]["dupe_of"] );
                           fieldList[k]->pCustomFieldValues[0]->pValueId = 
                              ValToStr( getResult["bugs"][0]["resolution"] );
                        } 
                        else
                        {
                           fieldList[k]->pCustomFieldValues[0]->pDisplayValue = 
                              ValToStr( getResult["bugs"][0]["resolution"] ) ;
                           fieldList[k]->pCustomFieldValues[0]->pValueId = 
                              ValToStr( getResult["bugs"][0]["resolution"] ) ;
                        }
                        fieldList[k]->pCustomFieldValues[0]->bSelected = true;
                        k++;
                        i++;
                     }

                     // loop through the rest of the fields
                     for ( ; k < cfCount; i++, k++)
                     {
                        fieldList[k]->pDisplayName = ValToStr(fieldResult["fields"][i]["display_name"]);
                        fieldList[k]->pFieldId = ValToStr(fieldNames["names"][i]);
                        fieldList[k]->eType = fieldTypes[i];
                        fieldList[k]->numOfFieldValues = 1;
                        fieldList[k]->pCustomFieldValues = AllocCustomFieldValueList(1); 
                        if (fieldList[k]->eType == eDateTime)
                        {
                           fieldList[k]->pCustomFieldValues[0]->pDisplayValue = 
                              TimeToStr( getResult["bugs"][0][(std::string)(bugNames[i])]);
                           fieldList[k]->pCustomFieldValues[0]->pValueId = 
                              TimeToStr( getResult["bugs"][0][(std::string)(bugNames[i])]);
                           fieldList[k]->pCustomFieldValues[0]->bSelected = true;
                        }
                        else
                        {
                           if (getResult["bugs"][0][(std::string)(bugNames[i])].getType() != 7)
                           {
                              fieldList[k]->pCustomFieldValues[0]->pDisplayValue =
                                 ValToStr(getResult["bugs"][0][(std::string)(bugNames[i])]);
                              fieldList[k]->pCustomFieldValues[0]->pValueId =
                                 ValToStr(getResult["bugs"][0][(std::string)(bugNames[i])]);
                              fieldList[k]->pCustomFieldValues[0]->bSelected = true;
                           }
                        }
                     }
                  }
                  else result = handleErrors(result, "Bug Fields", fieldResult, pErrorString);
               }
               // older version, use standard names
               else 
               {
                  std::string displayNames[] = {"Resolution","Priority","Severity","Alias",
                     "Duplicate","Reported","Modified","Assigned To"};

                  // if bug is closed, set resolution
                  if (!getResult["bugs"][0]["is_open"])
                  {
                     fieldList[k]->pDisplayName = CopyAndAllocStr("Resolution");
                     fieldList[k]->pFieldId = ValToStr(fieldNames["names"][0]);
                     fieldList[k]->eType = fieldTypes[0];
                     fieldList[k]->numOfFieldValues = 1;
                     fieldList[k]->pCustomFieldValues = AllocCustomFieldValueList(1); 
                     // if resolution is duplicate, add the dupe #
                     if( (std::string)getResult["bugs"][0]["resolution"] == "DUPLICATE")
                     {
                        fieldList[k]->pCustomFieldValues[0]->pDisplayValue = 
                           ValToStr( (std::string)getResult["bugs"][0]["resolution"] + 
                           " of bug # " + (std::string)getResult["bugs"][0]["dupe_of"] );
                        fieldList[k]->pCustomFieldValues[0]->pValueId = 
                           ValToStr( getResult["bugs"][0]["resolution"] );
                     } 
                     else
                     {
                        fieldList[k]->pCustomFieldValues[0]->pDisplayValue = 
                           ValToStr( getResult["bugs"][0]["resolution"] ) ;
                        fieldList[k]->pCustomFieldValues[0]->pValueId = 
                           ValToStr( getResult["bugs"][0]["resolution"] ) ;
                     }
                     fieldList[k]->pCustomFieldValues[0]->bSelected = true;
                     k++;
                     i++;
                  }

                  // loop through the rest of the fields
                  for ( ; i<cfCount; k++, i++)
                  {
                     fieldList[k]->pDisplayName = CopyAndAllocStr(displayNames[i].c_str());
                     fieldList[k]->pFieldId = ValToStr(fieldNames["names"][i]);
                     fieldList[k]->eType = fieldTypes[i];
                     fieldList[k]->numOfFieldValues = 1;
                     fieldList[k]->pCustomFieldValues = AllocCustomFieldValueList(1);
                     if (fieldList[k]->eType == eDateTime)
                     {
                        fieldList[k]->pCustomFieldValues[0]->pDisplayValue = 
                           TimeToStr( getResult["bugs"][0][(std::string)(bugNames[i])]);
                        fieldList[k]->pCustomFieldValues[0]->pValueId = 
                           TimeToStr( getResult["bugs"][0][(std::string)(bugNames[i])]);
                        fieldList[k]->pCustomFieldValues[0]->bSelected = true;
                     }
                     else
                     {
                        fieldList[k]->pCustomFieldValues[0]->pDisplayValue = 
                           ValToStr( getResult["bugs"][0][(std::string)(bugNames[i])]);
                        fieldList[k]->pCustomFieldValues[0]->pValueId = 
                           ValToStr( getResult["bugs"][0][(std::string)(bugNames[i])]);
                        fieldList[k]->pCustomFieldValues[0]->bSelected = true;
                     }
                  }
               }
               // assign pointers
               *customFields = fieldList;
               *numOfCustomFields = cfCount;
               *pIssue = myIssue;
            }
            else result = handleErrors(result, "Get Comments", comResult, pErrorString);
         }
         else result = handleErrors(result, "Get Bugs", getResult, pErrorString);
      }
      else result = handleErrors(result, "Bugzilla Version", verResult, pErrorString);
   }    
   return result;
}

// Gets the required workflow for an issue
IPExport SSCMIPResult GetFixWorkflow(const SSCMIPConnectionContext *context,
                                     const char *pIssueIds[],
                                     uint numOfIssues,
                                     SSCMIPCustomField **customFields[],
                                     uint *numOfCustomFields,
                                     char **pErrorString)
{
   uint cfCount = 2;
   SSCMIPCustomField** fieldList = AllocCustomFieldList(cfCount);

   fieldList[0]->pDisplayName = CopyAndAllocStr("Hours Worked");
   fieldList[0]->pFieldId = CopyAndAllocStr("work_time");
   fieldList[0]->eType = eTextfield;
   fieldList[0]->numOfFieldValues = 0;
   fieldList[0]->pCustomFieldValues = NULL;

   fieldList[1]->pDisplayName = CopyAndAllocStr("Comment");
   fieldList[1]->pFieldId = CopyAndAllocStr("comment");
   fieldList[1]->eType =eTextbody;
   fieldList[1]->numOfFieldValues = 0;
   fieldList[1]->pCustomFieldValues = NULL;

   *customFields = fieldList;
   *numOfCustomFields = cfCount;
   
   *pErrorString = NULL;
   return SSCMIP_OK;
}

// Makes an attachment to an issue
IPExport SSCMIPResult AttachToIssues(const SSCMIPConnectionContext *context,
                                     const SSCMIPIssueAttachment *issueAttachments[],
                                     uint numOfAttachments,
                                     SSCMIPFailedIssueAttachment **failedAttachments[],
                                     uint *numOfFailedAttachments,
                                     char **pErrorString)
{
   XmlRpcValue connection, bzParams, params, bzResult, scmserver;
   uint commentSize=0, counter, failNum = 0;
   std::string textFile, line;
   SSCMIPResult result =SSCMIP_OK;
   *pErrorString = NULL;
   *failedAttachments = NULL;
   char* commentBuffer;
   SSCMIPFailedIssueAttachment **localFailedAttachments = NULL;

   // get server info from config file 
   ifstream connectionFile(configFilePath.c_str());
   if (connectionFile.is_open())
   {
      while (!connectionFile.eof()) 
      {
         getline(connectionFile, line);
         textFile += line;
      }
      connectionFile.close();
      result = scmserver.serverInfo(textFile);
   }

   if (!result)
   {
      for (uint i=0; i<numOfAttachments; i++)
      {
         result = getConnectionInfo(&connection, issueAttachments[i]->pConnection, pErrorString);
         if (!result)
         {
            // setup XmlRpcClient according to selected connection
            XmlRpcClient Client (ValToStr(connection["server_url"]), connection["port"], 
               ValToStr(connection["web_service_extension"]));
            Client.setCookieFilePath(cookieFilePath);

            // allocate a c string for the comment
            uint j = 0, k =0;
            bool fix = false;
            if (issueAttachments[i]->numOfWorkflows == 2)
            {
               fix = true;
               commentSize += 15 + strlen(issueAttachments[i]->pCustomWorkflows[1]->pCustomFieldValues[0]->pDisplayValue);
            }
            do 
            {
               for( ; j < issueAttachments[i]->numOfFiles && commentSize <= 65530; j++)
               {
                  commentSize += 115 + strlen(issueAttachments[i]->pFiles[j]->pMainline) + 
                     2*strlen(issueAttachments[i]->pFiles[j]->pBranch) + 
                     2*strlen(issueAttachments[i]->pFiles[j]->pRepoPath) +
                     2*strlen(issueAttachments[i]->pFiles[j]->pFileName) +
                     scmserver["ip_address"].size();
               }
               if ( commentSize >= 65530 )
               {
                  commentSize -= (115 + strlen(issueAttachments[i]->pFiles[j]->pMainline) + 
                     2*strlen(issueAttachments[i]->pFiles[j]->pBranch) + 
                     2*strlen(issueAttachments[i]->pFiles[j]->pRepoPath) +
                     2*strlen(issueAttachments[i]->pFiles[j]->pFileName) +
                     scmserver["ip_address"].size());
               }
               commentBuffer = new char[commentSize];
               // loop through attachments, append info to comment
               counter = 0;
               // if fixing, add a fix comment, other fields
               if (fix) 
               {
#ifdef WIN32
                  counter += sprintf_s(commentBuffer + counter, commentSize-counter, "Fix Comment: %s\n\n", 
                     issueAttachments[i]->pCustomWorkflows[1]->pCustomFieldValues[0]->pDisplayValue);
#else
                  counter += sprintf(commentBuffer + counter, "Fix Comment: %s\n\n", 
                     issueAttachments[i]->pCustomWorkflows[1]->pCustomFieldValues[0]->pDisplayValue);
#endif
                  bzParams["work_time"] = issueAttachments[i]->pCustomWorkflows[0]->
                     pCustomFieldValues[0]->pDisplayValue;
               }
               for(; k < j; k++)
               {
      #ifdef WIN32
                  counter += sprintf_s(commentBuffer + counter, commentSize-counter,
                     "Surround SCM Attachment\nMainline: %s\nBranch: %s\nRepository: %s\nFile: %s\nVersion: %d\nsscm://%s:%d//%s//%s/%s\n\n", 
                     issueAttachments[i]->pFiles[k]->pMainline, 
                     issueAttachments[i]->pFiles[k]->pBranch, 
                     issueAttachments[i]->pFiles[k]->pRepoPath, 
                     issueAttachments[i]->pFiles[k]->pFileName, 
                     issueAttachments[i]->pFiles[k]->version,
                     ValToStr(scmserver["ip_address"]), 
                     (int) scmserver["port"],
                     issueAttachments[i]->pFiles[k]->pBranch, 
                     issueAttachments[i]->pFiles[k]->pRepoPath, 
                     issueAttachments[i]->pFiles[k]->pFileName );
      #else
                  counter += sprintf(commentBuffer + counter,
                     "Surround SCM Attachment\nMainline: %s\nBranch: %s\nRepository: %s\nFile: %s\nVersion: %d\nsscm://%s:%d//%s//%s/%s\n\n", 
                     issueAttachments[i]->pFiles[k]->pMainline, 
                     issueAttachments[i]->pFiles[k]->pBranch, 
                     issueAttachments[i]->pFiles[k]->pRepoPath, 
                     issueAttachments[i]->pFiles[k]->pFileName, 
                     issueAttachments[i]->pFiles[k]->version,
                     ValToStr(scmserver["ip_address"]), (int) scmserver["port"],
                     issueAttachments[i]->pFiles[k]->pBranch, 
                     issueAttachments[i]->pFiles[k]->pRepoPath, 
                     issueAttachments[i]->pFiles[k]->pFileName );
      #endif
               }

               bzParams["comment"] = commentBuffer;
               bzParams["id"] = issueAttachments[i]->pIssueId;
               bzParams["token"] = currentToken;
               params[0]=bzParams;

               // add the comment
               result = bzExecute(&Client, context, "Bug.add_comment", params, &bzResult);

               if (result || bzResult.hasMember("faultCode"))
               {
                  result = handleErrors(result, "Add Comment", bzResult, pErrorString);

                  SSCMIPFailedIssueAttachment** newFails = AllocFailedIssueAttachment(failNum+1);
                  
                  // If there are existing attachments, move their contents to the new list
                  if(localFailedAttachments != NULL)
                  {
                     for (uint q = 0; q < failNum; q++)
                     {
                        newFails[q]->pConnection = localFailedAttachments[q]->pConnection;
                        newFails[q]->pIssueId = localFailedAttachments[q]->pIssueId;
                        newFails[q]->pErrorMsg = localFailedAttachments[q]->pErrorMsg;
                     }
                  }
                  newFails[failNum]->pConnection = CopyAndAllocStr(issueAttachments[i]->pConnection);
                  newFails[failNum]->pIssueId = CopyAndAllocStr(issueAttachments[i]->pIssueId);
                  newFails[failNum]->pErrorMsg = *pErrorString;
                  
                  if(localFailedAttachments != NULL)
                  {
                     // We just delete the structures and the list, the strings in the structures were moved to newFails
                     for(uint q = 0; q < failNum; q++)
                     {
                        if(localFailedAttachments[q])
                           delete localFailedAttachments[q];
                     }
                     delete[] localFailedAttachments;
                     localFailedAttachments = NULL;
                  }
                  
                  failNum ++;
                  localFailedAttachments = newFails;
                  result = SSCMIP_OK;
                  *pErrorString = NULL;
               }
               fix = false;
               FreeString(commentBuffer);
               bzParams.clear();
               commentSize = 0;
            } while (j < issueAttachments[i]->numOfFiles);
         }
         else 
         {
            *pErrorString = CopyAndAllocStr("Error getting connection info from BugzillaConnections.txt");
            return SSCMIP_ERROR;
         }
      }
   }
   else
   {
      *pErrorString = CopyAndAllocStr("Error getting server info from BugzillaConnections.txt");
      return SSCMIP_ERROR;
   }
   *numOfFailedAttachments = failNum;
   *failedAttachments = localFailedAttachments;
   return result;
}

// Validates the user's workflow for an issue. 
// The errorString will only be displayed if workflowIsValid is true
IPExport SSCMIPResult ValidateWorkflow(const SSCMIPConnectionContext *context,
                                       const char *pIssueIds[],
                                       uint numOfIssues,
                                       const SSCMIPCustomField *customFields[], 
                                       uint numOfCustomFields,
                                       BOOL *workflowIsValid,
                                       char **pErrorString)
{
   SSCMIPResult result = SSCMIP_OK;
   char* end;
   BOOL valid = 1;
   if (numOfCustomFields != 2 )
   {
      *pErrorString = CopyAndAllocStr("Wrong Number of Fields");
      valid = 0;
   }
   else if ( strtol(customFields[0]->pCustomFieldValues[0]->pDisplayValue, &end, 10) > 99999 )
   {
      *pErrorString = CopyAndAllocStr("Too Many Hours Worked");
      valid = 0;
   }
   else if ( strlen(customFields[1]->pCustomFieldValues[0]->pDisplayValue) > 60000 )
   {
      *pErrorString = CopyAndAllocStr("Comment must be less than 60,000 characters");
      valid = 0;
   }
   *workflowIsValid = valid;
   return result;
}

/////////////////////////////////////////////////////
// Free Memory Functions
/////////////////////////////////////////////////////

IPExport void FreeString(char *str)
{
   if(str != NULL)
      delete[] str;
}

IPExport void FreeStringList(char *list[], uint numOfStrings)
{
   if(!list)
      return;
   
   uint i;
   for(i = 0; i < numOfStrings; i++)
      delete[] list[i];
   delete[] list;
}

IPExport void FreeConnectionList(SSCMIPConnectionDetails *connections[], 
                                 uint numOfConnections)
{
   if(!connections)
      return;
   
   SSCMIPConnectionDetails *currentConnection = NULL;
   uint i;
   
   for(i = 0; i < numOfConnections; i++)
   {
      currentConnection = connections[i];
      
      FreeString(currentConnection->pConnectionName);
      
      delete currentConnection;
   }
   
   delete[] connections;
}

IPExport void FreeCustomFieldsList(struct SSCMIPCustomField *customFields[], 
                                   uint numOfCustomFields)
{
   if(!customFields)
      return;
   
   SSCMIPCustomField *currentCustomField = NULL;
   SSCMIPCustomFieldValue *currentCustomFieldValue = NULL;
   
   uint i, j;
   
   for(i = 0; i < numOfCustomFields; i++)
   {
      currentCustomField = customFields[i];
      
      FreeString(currentCustomField->pDisplayName);
      FreeString(currentCustomField->pFieldId);
      
      for(j = 0; j < currentCustomField->numOfFieldValues; j++)
      {
         currentCustomFieldValue = currentCustomField->pCustomFieldValues[j];
         
         FreeString(currentCustomFieldValue->pDisplayValue);
         FreeString(currentCustomFieldValue->pValueId);
         
         delete currentCustomFieldValue;
      }
      
      delete[] currentCustomField->pCustomFieldValues;
      delete currentCustomField;
   }
   
   delete[] customFields;
}

IPExport void FreeIssue(SSCMIPIssue *issue)
{
   if(!issue)
      return;
   
   FreeString(issue->pIssueId);
   FreeString(issue->pDescription);
   FreeString(issue->pStatus);
   FreeString(issue->pSummary);
   FreeString(issue->pType);
   
   delete issue;
}

IPExport void FreeIssueList(SSCMIPIssue *issues[], uint numOfIssues)
{
   if(!issues)
      return;
   
   uint i;
   for(i = 0; i < numOfIssues; i++)
   {
      if(issues[i])
         FreeIssue(issues[i]);
   }
   
   delete[] issues;
}

IPExport void FreeFailedAttachmentList(SSCMIPFailedIssueAttachment *failedAttachments[], 
                                       uint numOfFailedAttachments)
{
   if(!failedAttachments)
      return;
   
   SSCMIPFailedIssueAttachment *pCurrent = NULL;
   
   for(uint i = 0; i < numOfFailedAttachments; i++)
   {
      pCurrent = failedAttachments[i];
      
      FreeString(pCurrent->pConnection);
      FreeString(pCurrent->pErrorMsg);
      FreeString(pCurrent->pIssueId);
      
      delete pCurrent;
   }
   
   delete[] failedAttachments;
}
 /////////////////////////////////////////////
 // Memory Allocate Functions
 ////////////////////////////////////////////

char* CopyAndAllocStr(const char *pStr)
{
   int length = (int)strlen(pStr)+1;
   char* pResult = new char[length];
   
#ifdef WIN32
   strcpy_s(pResult, length,  pStr);
#else
   strncpy(pResult, pStr, length);
#endif
   
   return pResult;
}

char* CopyAndAllocFormatStr(const char *pStr, int num)
{
   int length = (int)strlen(pStr)+10;
   char* pResult = new char[length];
   
#ifdef WIN32
   sprintf_s(pResult, length, pStr, num);
#else
   sprintf(pResult, pStr, num);
#endif
   return pResult;
}

char* CopyAndAllocFormatStr(const char *pStr,  const char* num)
{
   int length = (int)strlen(pStr)+(int)strlen(num) +1;
   char* pResult = new char[length];
   
#ifdef WIN32
   sprintf_s(pResult, length, pStr, num);
#else
   sprintf(pResult, pStr, num);
#endif
   return pResult;
}

SSCMIPConnectionDetails** AllocConnectionDetailsList(uint numOfItems)
{
   SSCMIPConnectionDetails **pResult = new SSCMIPConnectionDetails*[numOfItems];
   
   for(uint i = 0; i < numOfItems; i++)
   {
      pResult[i] = new SSCMIPConnectionDetails;
      
      pResult[i]->pConnectionName = NULL;
      pResult[i]->bConnectionSupportsBulkFixEvents = false;
   }
   
   return pResult;
}

SSCMIPCustomField** AllocCustomFieldList(uint numOfItems)
{
   SSCMIPCustomField **pResult = new SSCMIPCustomField*[numOfItems];
   
   for(uint i = 0; i < numOfItems; i++)
   {
      pResult[i] = new SSCMIPCustomField;
      
      pResult[i]->pDisplayName = NULL;
      pResult[i]->pFieldId = NULL;
      pResult[i]->pCustomFieldValues = NULL;
      pResult[i]->numOfFieldValues = 0;
      pResult[i]->eType = eTextbody;
   }
   
   return pResult;
}

SSCMIPCustomFieldValue** AllocCustomFieldValueList(uint numOfItems)
{
   SSCMIPCustomFieldValue **pResult = new SSCMIPCustomFieldValue*[numOfItems];
   
   for(uint i = 0; i < numOfItems; i++)
   {
      pResult[i] = new SSCMIPCustomFieldValue;
      
      pResult[i]->pDisplayValue = NULL;
      pResult[i]->pValueId = NULL;
      pResult[i]->bSelected = false;
   }
   
   return pResult;
}

SSCMIPIssue** AllocIssueList(uint numOfItems)
{
   SSCMIPIssue **pResult = new SSCMIPIssue*[numOfItems];
   
   for(uint i = 0; i < numOfItems; i++)
   {
      pResult[i] = new SSCMIPIssue;
      
      pResult[i]->pIssueId = NULL;
      pResult[i]->pDescription = NULL;
      pResult[i]->pStatus = NULL;
      pResult[i]->pSummary = NULL;
      pResult[i]->pType = NULL;
      pResult[i]->bUserCanFix = false;
      pResult[i]->bUserCanAttach = false;
   }
   
   return pResult;
}

SSCMIPFailedIssueAttachment** AllocFailedIssueAttachment(uint numOfItems)
{
   SSCMIPFailedIssueAttachment **pResult = new SSCMIPFailedIssueAttachment*[numOfItems];
   
   for(uint i = 0; i < numOfItems; i++)
   {
      pResult[i] = new SSCMIPFailedIssueAttachment;
      
      pResult[i]->pConnection = NULL;
      pResult[i]->pIssueId = NULL;
      pResult[i]->pErrorMsg = NULL;
   }
   
   return pResult;
}

/////////////////////////////////////////////////
// Additional Internal Functions
////////////////////////////////////////////////

// converts an XmlRpcValue TypeString to a c string
char* ValToStr(XmlRpcValue value)
{
   return CopyAndAllocStr(((std::string)(value)).c_str());
}

// appends to a c string
int append2 (int* i, char**buf, int size, const char* format, int num)
{
#ifdef WIN32
   if (num>9)
      return sprintf_s(*buf +*i, size -*i, format, num);
   else
   {
      *i += sprintf_s(*buf +*i, size -*i, "%d", 0);
      return sprintf_s(*buf +*i, size -*i, format, num);
   }
#else
   if (num>9)
      return sprintf(*buf +*i, format, num);
   else
   {
      *i += sprintf(*buf +*i, "%d", 0);
      return sprintf(*buf +*i, format, num);
   }
#endif

}

// converts XmlRpcValue TypeDateTime into a string
char* TimeToStr(XmlRpcValue value)
{
   int i, length = 30;
   struct tm timeVal = value;
   char* strVal = new char[length];
#ifdef WIN32
   i  = sprintf_s(strVal, length, "%d-", timeVal.tm_year);
#else
   i  = sprintf(strVal, "%d-", timeVal.tm_year);
#endif
   i += append2(&i, &strVal, length, "%d-", timeVal.tm_mon);
   i += append2(&i, &strVal, length, "%dT", timeVal.tm_mday);
   i += append2(&i, &strVal, length, "%d:", timeVal.tm_hour);
   i += append2(&i, &strVal, length, "%d:", timeVal.tm_min);
   i += append2(&i, &strVal, length, "%d", timeVal.tm_sec); 
   return strVal;
}

// logs in a user
SSCMIPResult bzLogin(XmlRpcClient c, 
                     const SSCMIPConnectionContext* info, 
                     XmlRpcValue* returned)
{
   XmlRpcValue bzParams, params;
   bzParams["login"] = info->pUsername;
   bzParams["password"] = info->pPassword;
   params[0]=bzParams;
   SSCMIPResult result = c.execute("User.login", params, *returned);
   if (!result && !returned->hasMember("faultCode"))
   {
      currentUser.clear();
      currentUser = info->pUsername;
      currentToken = (std::string)returned[0]["token"];
   }
   return result;
}

// validates cookies, logs in (if necessary) and executes a function
SSCMIPResult bzExecute(XmlRpcClient* c, 
                       const SSCMIPConnectionContext* info, 
                       const char* function, 
                       XmlRpcValue parameters, 
                       XmlRpcValue* sucess)
{
   SSCMIPResult result = SSCMIP_OK;
   BOOL fault = false;

   ifstream cookieFile;
   cookieFile.open(cookieFilePath.c_str(), ios::in);

   // check to see if we have cookies
   if (!cookieFile.is_open())
   {
      result = bzLogin(*c, info, sucess);
      fault = sucess->hasMember("faultCode");
   }
   // check to see if the cookies we have are expired, or the user has changed
   else if (expired(&cookieFile) || !rightUser(*info) || currentToken.empty())
   {
      cookieFile.close();
      remove(cookieFilePath.c_str());
      result = bzLogin(*c, info, sucess);
      fault = sucess->hasMember("faultCode");
   }
   // if we are logged in, execute the function
   if (!result && !fault) 
   {
      result = c->execute(function, parameters, *sucess);
   }

   return result;
}

// handles a failed execution, or server returned fault
SSCMIPResult handleErrors(SSCMIPResult status, 
                          const char* operation, 
                          XmlRpcValue result,  
                          char** errorString)
{
   switch (status)
   {
      case 0:
         if (result.hasMember("faultCode"))
         {
            std::string fString = result["faultString"];
            std::string fCode = result["faultCode"];
            *errorString = CopyAndAllocStr(fString.c_str());
            if (fCode == "Client") status = 1;
            else status = strtol(fCode.c_str(), NULL, 10);
         }
         break;
      case CONNECTION_FAILED:
         *errorString = CopyAndAllocFormatStr("%s failed to connect to Bugzilla server", operation);
         break;
      case REQUEST_FAILED:
         *errorString = CopyAndAllocFormatStr("%s failed to send request to Bugzilla server", operation);
         break;
      case RESPONSE_FAILED:
         *errorString = CopyAndAllocFormatStr("%s failed to receive response from Bugzilla server", operation);
         break;
   }
   return status;
}

// checks if a cookie is expired
BOOL expired ( ifstream* cookie )
{
   char* buffer = new char[500];
   char* month = new char[4];
   int date=0, mdays, year=0, hour=0, min=0;
   BOOL result;

   // get expiration date from cookie
   do {
      (*cookie).getline(buffer, 500, ';');
   } while ( !cookie->eof() && strncmp(buffer, " expires", 8) !=  0 );

   // if we reached the eof and the buffer is empty, then return false
   if (strncmp(buffer, " expires", 8) != 0)
      return false;

   // expires=Fri, 01-Jan-2038 00:00:00 GMT
   // extract values from string
#ifdef WIN32
   sscanf_s(buffer, "%*s %d  %*c %3c - %d %d : %d", &date, month, 4, &year, &hour, &min);
#else
   sscanf(buffer, "%*s %d  %*c %3c - %d %d : %d", &date, month, &year, &hour, &min);
#endif
   month[3] = '\0';

   // translate month into days since jan 1
   if ( !strcmp(month, "Jan") ) mdays = 0;
   else if ( !strcmp(month, "Feb") ) mdays = 31;
   else if ( !strcmp(month, "Mar") ) mdays = 59;
   else if ( !strcmp(month, "Apr") ) mdays = 90;
   else if ( !strcmp(month, "May") ) mdays = 120;
   else if ( !strcmp(month, "Jun") ) mdays = 151;
   else if ( !strcmp(month, "Jul") ) mdays = 181;
   else if ( !strcmp(month, "Aug") ) mdays = 212;
   else if ( !strcmp(month, "Sep") ) mdays = 243;
   else if ( !strcmp(month, "Oct") ) mdays = 273;
   else if ( !strcmp(month, "Nov") ) mdays = 304;
   else if ( !strcmp(month, "Dec") ) mdays = 334;
   else return true;

   // number of leap years since 1970
   double leaps = ceil((double)((year - 1972)/4));
   // number of seconds since 1970
   double secs = ( ( ( year*365 + leaps + mdays + date)*24 + hour)*60 + min )*60 ;
   // current number of seconds since 1970
   time_t now = time(NULL);
   // compare
   if ((secs - now) > 3600) result = false;
   else result = true;
   return result;   
}

// reads BugzillaConnections.xml, puts values in an XmlRpcValue array of structs
// reads BugzillaConnections.xml, puts values in an XmlRpcValue array of structs
SSCMIPResult readBugzillaConnections(XmlRpcValue* myList, 
                                     char** errorString)
{
   SSCMIPResult result = SSCMIP_OK;
   std::string connectionListXml, line;

   ifstream connectionFile;
   connectionFile.open(configFilePath.c_str());
   if (connectionFile.is_open())
   {
      // read from text file to string
      while (!connectionFile.eof()) 
      {
         getline(connectionFile, line);
         connectionListXml += line;
      }
      connectionFile.close();
      // connectionList converts Xml into array of structs

      result = myList->connectionList(connectionListXml);

      if (result)
      {
         if(myList->getType() != XmlRpcValue::TypeArray)
         {
            *errorString = CopyAndAllocStr("BugzillaConnections.txt is formatted incorrectly");
         }
         else
         {
            for (int i=0; i<myList->size(); i++)
            {
               if ((*myList)[i].hasMember("fault"))
               {
                  *errorString = ValToStr(myList[i]["fault"]);
               }
            }
         }
      }
   }
   else
   {
      *errorString = CopyAndAllocStr("BugzillaConnections.txt does not exist");
      result = SSCMIP_ERROR;
   }

   return result;
}

// gets details for a specific connection
SSCMIPResult getConnectionInfo(XmlRpcValue* info, 
                               const char* connectionName, 
                               char** errorString)
{
   XmlRpcValue list;
   if (!readBugzillaConnections(&list, errorString))
   {
      for (int i=0; i<list.size(); i++)
      {
         if ( strcmp( ValToStr(list[i]["name"]), connectionName) == 0)
         {
            *info = list[i];
            *errorString = NULL;
            return SSCMIP_OK;
         }
      }
   }
   *errorString = CopyAndAllocStr("Connection Not Found");
   return SSCMIP_ERROR;
}

// checks to see if the user has changed and we need to log in again
BOOL rightUser(SSCMIPConnectionContext context)
{
   if ( currentUser.compare(context.pUsername) == 0 ) return true;
   else return false;
}




